@@ -1,4 +1,4 @@
-# Copyright 2017-2020 R.A. Stern
+# Copyright 2017-2021 R.A. Stern
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -377,10 +377,11 @@ class VersionSpec:
         self.allow_snapshot = snapshot
         self.cmp_base = cmp_base
 
-        try:
-            self.versions.sort()
-        except AttributeError:
-            raise VMTFormatError('Invalid input format')
+        if self.versions:
+            try:
+                self.versions.sort()
+            except AttributeError:
+                raise VMTFormatError('Invalid input format')
 
     @staticmethod
     def str_to_ver(string):
@@ -407,11 +408,12 @@ class VersionSpec:
 
     @staticmethod
     def _check(current, versions, required=True, warn=True):
-        for v in versions:
-            res = VersionSpec.cmp_ver(current, v)
+        if versions:
+            for v in versions:
+                res = VersionSpec.cmp_ver(current, v)
 
-            if (res >= 0 and v[-1] == '+') or res == 0:
-                return True
+                if (res >= 0 and v[-1] == '+') or res == 0:
+                    return True
 
         if required:
             raise VMTVersionError('Required version not met')
@@ -559,7 +561,7 @@ class Pager:
 
         if 'cursor' in partial:
             self.__resource, self.__query = partial.split('?', 1)
-            self.__query = re.sub(r'(?<=\?|&)cursor=([\d]+)', f"cursor={self.__next}", self.__query)
+            self.__query = re.sub(r'(?<=\?|&)cursor=(.+?)(?![^&])', f"cursor={self.__next}", self.__query)
         else:
             try:
                 self.__resource, self.__query = partial.split('?', 1)
@@ -657,7 +659,7 @@ class Connection:
     Each of these override the connection global property, and will be safely
     ignored if the endpoint does not support, or does not require paging the
     results. Additionally, you may pass :py:class:`requests.Request` keyword
-    arguments to each call if required (e.g. `timeout <https://requests.readthedocs.io/en/master/user/quickstart/#timeouts>`_).
+    arguments to each call if required (e.g. `timeout <https://docs.python-requests.org/en/master/user/quickstart/#timeouts>`_).
     Care should be taken, as some parameters will break *vmt-connect* calls if they
     conflict with existing headers, or alter expected results.
 
@@ -857,8 +859,12 @@ class Connection:
 
     def __login(self):
         u, p = (base64.b64decode(self.__basic_auth)).decode().split(':', maxsplit=1)
-        body = {'username': (None, u), 'password': (None, p)}
-        self.request('login', 'POST', disable_hateoas=False, content_type=None, files=body, allow_redirects=False)
+        self.request('login',
+                     'POST',
+                     disable_hateoas=False,
+                     content_type=None,
+                     files={'username': (None, u), 'password': (None, p)},
+                     allow_redirects=False)
 
     def __use_session(self, value):
         if value:
@@ -1067,7 +1073,7 @@ class Connection:
             except KeyError:
                 pass
 
-        if isinstance(query, str):
+        if query and isinstance(query, str):
             msg = 'Query parameters should be passed in as a dictionary.'
             warnings.warn(msg, DeprecationWarning)
 
@@ -1099,6 +1105,8 @@ class Connection:
             res = util.filter_copy(self.last_response.content.decode(),
                                    filter,
                                    use_float=filter_float)
+        elif method == "DELETE" and len(self.last_response.content) == 0:
+            res = list()
         else:
             res = self.last_response.json()
 
@@ -1172,7 +1180,7 @@ class Connection:
             elif id == '__group_members':
                 self.__inventory_cache[id]['data'] = self.request(f'groups/{uuid}/members', **kwargs)
             else:
-                self.__inventory_cache[id]['data'] = self.request(f'markets/{id}/entities', fetch_all=True, **kwargs)
+                self.__inventory_cache[id]['data'] = self.request(f'markets/{uuid}/entities', fetch_all=True, **kwargs)
 
             delta = datetime.timedelta(seconds=self.__inventory_cache_timeout)
             self.__inventory_cache[id]['expires'] = datetime.datetime.now() + delta
@@ -1289,12 +1297,6 @@ class Connection:
         Returns:
             A list of entities in :obj:`dict` form.
 
-
-        Notes:
-            **type** filtering is performed locally and is not compatible with
-            responses that return a :py:class:`~Pager` object. Therefore, if you
-            attempt to request a :py:class:`~Pager` response, **type** will be
-            ignored.
         """
         query = {}
 
@@ -1319,7 +1321,7 @@ class Connection:
             else:
                 entities = self.request(path, method='GET', query=query, **kwargs)
 
-        if type and isinstance(entities, Pager):
+        if type and isinstance(entities, list):
             return [deepcopy(x) for x in entities if x['className'] == type]
 
         return entities
@@ -1334,6 +1336,8 @@ class Connection:
         Returns:
             A list of virtual machines in :obj:`dict` form.
         """
+        if uuid is None:
+            kwargs["fetch_all"] = True
         return self.get_entities('VirtualMachine', uuid=uuid, market=market, **kwargs)
 
     def get_physicalmachines(self, uuid=None, market='Market', **kwargs):
@@ -1346,6 +1350,8 @@ class Connection:
         Returns:
             A list of hosts in :obj:`dict` form.
         """
+        if uuid is None:
+            kwargs["fetch_all"] = True
         return self.get_entities('PhysicalMachine', uuid=uuid, market=market, **kwargs)
 
     def get_datacenters(self, uuid=None, market='Market', **kwargs):
@@ -1358,6 +1364,8 @@ class Connection:
         Returns:
             A list of datacenters in :obj:`dict` form.
         """
+        if uuid is None:
+            kwargs["fetch_all"] = True
         return self.get_entities('DataCenter', uuid=uuid, market=market, **kwargs)
 
     def get_datastores(self, uuid=None, market='Market', **kwargs):
@@ -1370,6 +1378,8 @@ class Connection:
         Returns:
             A list of datastores in :obj:`dict` form.
         """
+        if uuid is None:
+            kwargs["fetch_all"] = True
         return self.get_entities('Storage', uuid=uuid, market=market, **kwargs)
 
     def get_clusters(self, uuid=None, cache=False, **kwargs):
@@ -1411,6 +1421,17 @@ class Connection:
                         if uuid == vm['uuid']:
                             return c
 
+    def get_entity_actions(self, uuid, **kwargs):
+        """Returns a list of entity actions.
+
+        Args:
+            uuid (str): Entity UUID.
+
+        Returns:
+            A list containing all actions for the given the entity.
+        """
+        return self.request(f'entities/{uuid}/actions', **kwargs)
+
     def get_entity_groups(self, uuid, **kwargs):
         """Returns a list of groups the entity belongs to.
 
@@ -1511,7 +1532,7 @@ class Connection:
 
         return self.request('groups', uuid=uuid, **kwargs)
 
-    def get_group_actions(self, uuid=None, **kwargs):
+    def get_group_actions(self, uuid, **kwargs):
         """Returns a list of group actions.
 
         Args:
@@ -1531,11 +1552,10 @@ class Connection:
         Returns:
             A list containing the group in :obj:`dict` form.
         """
-        groups = self.get_groups(**kwargs)
+        groups = self.search(types=["Group"], q=name, **kwargs)
 
-        for grp in groups:
-            if grp['displayName'] == name:
-                return [grp]
+        if isinstance(groups, list) and len(groups) > 0:
+            return [groups[0]]
 
         return None
 
@@ -1654,6 +1674,17 @@ class Connection:
         """
         return self.request('targets', uuid=uuid, **kwargs)
 
+    def get_target_actions(self, uuid, **kwargs):
+        """Returns a list actions on a target.
+
+        Args:
+            uuid (str): Entity UUID.
+
+        Returns:
+            A list containing all actions for entities of the given the target.
+        """
+        return self.request(f'targets/{uuid}/actions', **kwargs)
+
     def get_target_for_entity(self, uuid=None, name=None,
                               type='VirtualMachine', **kwargs):
         """Returns a list of templates.
@@ -1835,6 +1866,7 @@ class Connection:
             entity_types (list, optional): Member entity types filter.
             probe_types (list, optional): Target probe type filter.
             regex (bool, optional): Flag for regex query string searching.
+            query_type (str, optional): Specifies the type of query.
             Set
             uuid (str): UUID of an object to lookup.
             Set
@@ -1861,13 +1893,16 @@ class Connection:
         query = {}
         remove = []
         args = ['q', 'types', 'scopes', 'state', 'environment_type', 'group_type',
-        'detail_type', 'entity_types', 'regex', 'probe_types']
+        'detail_type', 'entity_types', 'regex', 'probe_types', 'query_type']
 
         for k in args:
             v = kwargs.get(k)
             if v is not None:
                 if k in ['types', 'scopes', 'entity_types', 'probe_types']:
                     query[k] = ','.join(v)
+                elif k == 'query_type':
+                    if v.upper() in ['REGEX', 'EXACT', 'CONTAINS']:
+                        query[k] = v
                 else:
                     query[k] = v
 
@@ -1875,7 +1910,6 @@ class Connection:
 
         for x in remove:
             del kwargs[x]
-
         return self.request('search', query=query, **kwargs)
 
     def search_by_name(self, name, type=None, case_sensitive=False,
