@@ -13,11 +13,19 @@
 from collections import defaultdict
 from copy import deepcopy
 from decimal import Decimal
-from io import StringIO
+from io import StringIO, BytesIO
 import re
 
 
 
+_template_resources = [
+    'computeResources',
+    'infrastructureResources',
+    'storageResources'
+]
+
+
+
 def enumerate_stats(data, entity=None, period=None, stat=None):
     """Enumerates stats endpoint results
 
@@ -62,6 +70,49 @@ def enumerate_stats(data, entity=None, period=None, stat=None):
                 yield v2['date'], v3
 
 
+def enumerate_template_resources(data, restype=None, res=None):
+    """Enumerates template commodities
+
+    Provides an iterator for more intuitive and cleaner parsing of nested
+    template resource data. Each iteration returns a tuple containing the
+    resource type, as well as the next resource entry as a dictionary.
+
+    Args:
+        data (list): Template data to parse.
+        restype (function, optional): Optional resource type level filter function.
+        res (function, optional): Optional resource level filter function.
+
+    Notes:
+        Filter functions must return ``True``, to continue processing, or ``False``
+        to skip processing the current level element.
+
+    Examples:
+        .. code-block:: python
+
+            # filter stats for a single type
+            desired_res = 'computeResources'
+            enumerate_template_resources(data, restype=lambda x: x == desired_res)
+
+            # filter specific stats for all IDs
+            blacklist = ['diskConsumedFactor']
+            enumerate_template_resources(data, res=lambda x: x['name'] not in blacklist)
+    """
+    if isinstance(data, list):
+        data = data[0]
+
+    for r in _template_resources:
+        if restype is not None and not restype(r):
+            continue
+
+        try:
+            for k1, v1 in enumerate(data[r][0]['stats']):
+                if res is not None and not res(v1):
+                    continue
+                yield r, v1
+        except KeyError:
+            pass
+
+
 def unit_cast(value, ufrom, uto, factor, unit_list, precision=False):
     offset = unit_list.index(uto) - unit_list.index(ufrom)
     chg = Decimal(pow(factor, abs(offset)))
@@ -186,7 +237,8 @@ def filter_copy(source, filter, size=500, use_float=False):
     out = [None] * size
     idx = 0
 
-    for x in ijson.items(StringIO(source), 'item', use_float=use_float):
+    # to remove ijson BytesIO warning
+    for x in ijson.items(BytesIO(source.encode('utf-8')), 'item', use_float=use_float):
         if idx >= len(out):
             out.extend([None] * size)
 
